import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;

import latch.Latch;
import latch.LatchSupport;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.dbi.INList;
import com.sleepycat.je.evictor.Evictor;
import com.sleepycat.je.tree.IN;

public privileged aspect INListExtension {

	/* If both latches are acquired, major must always be acquired first. */
	private Latch INList.majorLatch;

	private Latch INList.minorLatch;

	private Set INList.addedINs = null;

	// Lck hurra homogen
	after(INList l, EnvironmentImpl envImpl) : 
		((execution(INList.new(EnvironmentImpl)) && args(envImpl)) ||
		(execution(INList.new(INList,EnvironmentImpl)) && args(INList, envImpl))) 
		&& target(l) {
		l.majorLatch = LatchSupport.makeLatch(l.DEBUG_NAME + " Major Latch",
				envImpl);
		l.minorLatch = LatchSupport.makeLatch(l.DEBUG_NAME + " Minor Latch",
				envImpl);
	}

	void around(INList l) throws DatabaseException : 
		execution(void INList.add(IN)) && target(l) {
		boolean enteredWithLatchHeld = l.majorLatch.isOwner();
		boolean addToMajor = true;
		try {
			if (enteredWithLatchHeld) {

				/*
				 * Don't try to acquire the major latch twice, that's not
				 * supported. If we do hold the latch, don't modify the major
				 * list, we may be faulting in an IN while iterating over the
				 * list from the evictor.
				 */
				addToMajor = false;
			} else {
				if (!(l.majorLatch.acquireNoWait())) {
					/* We couldn't acquire the latch. */
					addToMajor = false;
				}
			}

			if (addToMajor) {
				proceed(l);
			} else {
				l.minorLatch.acquire();
				try {
					proceed(l);
				} finally {
					l.minorLatch.release();
				}

				/*
				 * The holder of the majorLatch may have released it. If so, try
				 * to put the minor list into the major list so no INs are
				 * orphaned.
				 */
				if (!enteredWithLatchHeld) {
					if (l.majorLatch.acquireNoWait()) {
						try {
							l.latchMinorAndDumpAddedINs();
						} finally {
							l.releaseMajorLatch();
						}
					}
				}
			}
		} finally {
			if (addToMajor) {
				l.releaseMajorLatchIfHeld();
			}
		}
	}

	void around(INList l) throws DatabaseException :
		execution(void INList.remove(IN)) && target(l) && 
		!cflow(execution(long Evictor.evictIN(..))/* removeLatchAlreadyHeld */)
	{
		assert LatchSupport.countLatchesHeld() == 0;
		l.majorLatch.acquire();
		try {
			assert l.majorLatch.isOwner();
			proceed(l);
		} finally {
			l.releaseMajorLatch();
		}
	}

	boolean around(INList l, IN in) throws DatabaseException : 
		execution(boolean hook_doRemove(IN)) && this(l) && args(in) {
		boolean removeDone = proceed(l, in);
		if (!removeDone) {
			/* Look in addedINs set. */
			l.minorLatch.acquire();
			try {
				removeDone = l.addedINs.remove(in);
				l.dumpAddedINsIntoMajorSet();
			} finally {
				l.minorLatch.release();
			}
		}
		return removeDone;
	}

	pointcut assertOwner(INList l) : (  
			execution(SortedSet INList.tailSet(IN)) ||
			execution(IN INList.first()) ||
			execution(Iterator INList.iterator())
		) && target(l);

	before(INList l):assertOwner(l){
		assert l.majorLatch.isOwner();
	}

	void around(INList l) throws DatabaseException :
		execution(void INList.hook_doClear(SortedSet)) && this(l)
	{
		assert LatchSupport.countLatchesHeld() == 0;
		l.majorLatch.acquire();
		l.minorLatch.acquire();
		proceed(l);
		l.addedINs.clear();
		l.minorLatch.release();
		l.releaseMajorLatch();
	}

	/**
	 * The locking hierarchy is: 1. INList major latch. 2. IN latch. In other
	 * words, the INList major latch must be taken before any IN latches to
	 * avoid deadlock.
	 */
	public void INList.latchMajor() throws DatabaseException {

		assert LatchSupport.countLatchesHeld() == 0;
		majorLatch.acquire();
	}

	public void INList.releaseMajorLatchIfHeld() throws DatabaseException {

		if (majorLatch.isOwner()) {
			releaseMajorLatch();
		}
	}

	public void INList.releaseMajorLatch() throws DatabaseException {

		/*
		 * Before we release the major latch, take a look at addedINs and see if
		 * anything has been added to it while we held the major latch. If so,
		 * added it to ins.
		 */
		latchMinorAndDumpAddedINs();
		majorLatch.release();
	}

	private void INList.dumpAddedINsIntoMajorSet() {
		if (addedINs.size() > 0) {
			ins.addAll(addedINs);
			addedINs.clear();
		}
	}

	void INList.latchMinorAndDumpAddedINs() throws DatabaseException {

		latchMinor();
		try {
			dumpAddedINsIntoMajorSet();
		} finally {
			releaseMinorLatch();
		}
	}

	private void INList.latchMinor() throws DatabaseException {

		minorLatch.acquire();
	}

	private void INList.releaseMinorLatch() throws DatabaseException {

		minorLatch.release();
	}

}
