import latch.LatchSupport;
import latch.SharedLatch;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.cleaner.FileProcessor;
import com.sleepycat.je.dbi.DatabaseImpl;
import com.sleepycat.je.dbi.DbTree;
import com.sleepycat.je.recovery.RecoveryManager;
import com.sleepycat.je.tree.ChildReference;
import com.sleepycat.je.tree.IN;
import com.sleepycat.je.tree.Node;
import com.sleepycat.je.tree.Tree;
import com.sleepycat.je.tree.WithRootLatched;

public privileged aspect LatchedWalkerExtension {
	private SharedLatch Tree.rootLatch;

	pointcut executeWithRoot(Tree tree, WithRootLatched wrl) : 
		execution(IN Tree.withRoot(WithRootLatched)) && args(wrl) && target(tree);
	
	pointcut sharedWithRoot() :
		cflowbelow(
				execution(* FileProcessor.isRoot(..)) ||
				execution(* DbTree.getHighestLevel(..)));
	
	Object around(Tree tree, WithRootLatched wrl) throws DatabaseException : 
		executeWithRoot(tree,wrl) && sharedWithRoot() {
		try {
			tree.rootLatch.acquireShared();
			return proceed(tree,wrl);
		} finally {
			tree.rootLatch.release();
		}
	}
	
	Object around(Tree tree, WithRootLatched wrl) throws DatabaseException : 
		executeWithRoot(tree,wrl) && !sharedWithRoot() {
		try {
			tree.rootLatch.acquireExclusive();
			return proceed(tree,wrl);
		} finally {
			tree.rootLatch.release();
		}
	}
	

	before(Tree tree, DatabaseImpl database) : 
		execution(* Tree.init(..)) && target(tree) && args(database) {
		tree.rootLatch = LatchSupport.makeSharedLatch("RootLatch",
				(database != null) ? database.getDbEnvironment() : null);
	}

	pointcut setRootNotLatched(Tree tree) :
		call(void Tree.setRoot(ChildReference)) && target(tree) && cflowbelow(
				execution(* RecoveryManager.*(..))
				);
	
	before(Tree tree):setRootNotLatched(tree){
		assert tree.rootLatch.isWriteLockedByCurrentThread();
	}

	/* Caller is responsible for releasing rootLatch. */
	private SharedLatch Tree.RootChildReference.getRootLatch() {
		return Tree.this.rootLatch;
	}

	private void Tree.RootChildReference.assertLatch() {
		assert getRootLatch().isWriteLockedByCurrentThread();
	}

	public Node Tree.RootChildReference.fetchTarget(DatabaseImpl database, IN in)
			throws DatabaseException {

		if (getTarget() == null
				&& !Tree.this.rootLatch.isWriteLockedByCurrentThread()) {
			Tree.this.rootLatch.release();
			Tree.this.rootLatch.acquireExclusive();
		}

		return super.fetchTarget(database, in);
	}

	public void Tree.RootChildReference.setTarget(Node target) {
		super.setTarget(target);
	}

	public void Tree.RootChildReference.clearTarget() {
		super.clearTarget();
	}

	public void Tree.RootChildReference.setLsn(long lsn) {
		super.setLsn(lsn);
	}

	pointcut latchedRootChildMethods(Tree.RootChildReference rcr) : (
			execution(void Tree.RootChildReference.setTarget(..)) ||
			execution(void Tree.RootChildReference.clearTarget(..)) ||
			execution(void Tree.RootChildReference.setLsn(..)) 
		) && target(rcr);

	before(Tree.RootChildReference rcr) : latchedRootChildMethods(rcr) {
		rcr.assertLatch();
	}
}
