import java.nio.ByteBuffer;

import latch.Latch;
import latch.LatchNotHeldException;
import latch.LatchSupport;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.dbi.DatabaseImpl;
import com.sleepycat.je.dbi.EnvironmentImpl;
import com.sleepycat.je.tree.BIN;
import com.sleepycat.je.tree.BINDelta;
import com.sleepycat.je.tree.IN;

public aspect TreeLatchINExtension {
	private Latch IN.latch;

	/**
	 * Latch this node, optionally setting the generation.
	 */
	public void IN.latch(boolean updateGeneration) throws DatabaseException {

		if (updateGeneration) {
			setGeneration();
		}
		latch.acquire();
	}

	/**
	 * Latch this node if it is not latched by another thread, optionally
	 * setting the generation if the latch succeeds.
	 */
	public boolean IN.latchNoWait(boolean updateGeneration)
			throws DatabaseException {

		if (latch.acquireNoWait()) {
			if (updateGeneration) {
				setGeneration();
			}
			return true;
		} else {
			return false;
		}
	}

	/**
	 * Latch this node and set the generation.
	 */
	public void IN.latch() throws DatabaseException {

		latch(true);
	}

	/**
	 * Latch this node if it is not latched by another thread, and set the
	 * generation if the latch succeeds.
	 */
	public boolean IN.latchNoWait() throws DatabaseException {

		return latchNoWait(true);
	}

	/**
	 * Release the latch on this node.
	 */
	public void IN.releaseLatch() throws LatchNotHeldException {

		latch.release();
	}

	/**
	 * Release the latch on this node.
	 */
	public void IN.releaseLatchIfOwner() throws LatchNotHeldException {

		latch.releaseIfOwner();
	}

	/**
	 * @return true if this thread holds the IN's latch
	 */
	public boolean IN.isLatchOwner() {
		return latch.isOwner();
	}

	after(IN in):execution (void IN.init(DatabaseImpl, byte[],int,int)) && this(in) {
		DatabaseImpl databaseImpl = in.getDatabase();
		EnvironmentImpl env = (databaseImpl == null) ? null : databaseImpl
				.getDbEnvironment();
		in.latch = LatchSupport.makeLatch(in.shortClassName() + in.getNodeId(),
				env);
	}

	pointcut latchedMethods(IN in): (
			(execution(boolean IN.validateSubtreeBeforeDelete(int)) && within(IN)) ||
			execution(void IN.verify(byte[])) ||
			execution(boolean IN.isValidForDelete())
		) && this(in);

	Object around(IN in) throws DatabaseException : latchedMethods(in){
		boolean needToLatch = !in.isLatchOwner();
		try {
			if (needToLatch) {
				in.latch();
			}
			return proceed(in);
		} finally {
			if (needToLatch) {
				in.releaseLatchIfOwner();
			}
		}
	}

	pointcut assertLatched(IN in): (
			(execution(* IN.findParent(..)) && within(IN)) ||
			execution(* BIN.addCursor(..)) ||
			execution(* BIN.removeCursor(..)) ||
			execution(* BIN.adjustCursors(..)) ||
			execution(* BIN.adjustCursorsForInsert(..)) ||
			execution(* BIN.adjustCursorsForMutation(..)) ||
			execution(* BIN.evictLNs(..)) 
		) && this(in);

	before(IN in) : assertLatched(in){
		assert in.isLatchOwner();
	}

	boolean around(IN in, boolean updateGeneration) throws DatabaseException :
		execution(boolean IN.isSoughtNode(long, boolean)) && this(in) && args(long,updateGeneration) {
		in.latch(updateGeneration);
		boolean result = proceed(in, updateGeneration);
		if (result)
			in.releaseLatch();
		return result;
	}

	after(IN in):
		execution(void IN.readFromLog(ByteBuffer, byte)) 
		&& this(in) && within(IN) 
	{
		in.latch.setName(in.shortClassName() + in.getNodeId());
	}

	void around(IN newSibling) throws DatabaseException : 
		call(void IN.hook_splitInternal_work(IN,..)) && args(newSibling, ..){
		newSibling.latch();
		try {
			proceed(newSibling);
		} finally {
			newSibling.releaseLatch();
		}
	}

	// before(IN in, IN newParent) throws DatabaseException:
	// set(IN SearchResult.parent) && args(newParent) && if(newParent==null) &&
	// withincode(* IN.findParent(..)) && this(in) {
	// in.releaseLatch();
	// }
	before(IN in) throws DatabaseException : 
		execution(void IN.hook_releaseLatch()) && target(in) {
		in.releaseLatch();
	}

	before(IN newSibling) : execution(void BIN.adjustCursors(IN,int,int)) && within(BIN) && args(newSibling,int,int) {
		assert newSibling.isLatchOwner();
	}

	void around(BIN fullBIN) throws DatabaseException:
		execution(void BINDelta.hook_reconstitueBIN_internal(BIN)) && args(fullBIN){
		fullBIN.latch();
		proceed(fullBIN);
		fullBIN.releaseLatch();
	}

	before(IN in) throws DatabaseException : 
		execution(void IN.hook_releaseLatch()) && target(in) {
		in.releaseLatch();
	}

	boolean around(IN in, boolean b) throws DatabaseException : 
		execution(boolean IN.hook_latchNoWait(boolean)) && target(in) && args(b) {
		return in.latchNoWait(b);
	}
}
