package com.sleepycat.je;

import java.io.File;
import java.util.ArrayList;

import latch.LatchSupport;
import latch.SharedLatch;

import com.sleepycat.je.dbi.EnvironmentImpl;

/**
 * homogene pointcuts, damit kann man sich die vorherigen methoden eigentlich
 * sparen. dadurch schnerer sourcecode
 */
public privileged aspect TriggerLatchExtension {
	/*
	 * Used by Database to protect access to the trigger list. A single latch
	 * for all databases is used to prevent deadlocks.
	 */
	private SharedLatch EnvironmentImpl.triggerLatch;

	after(EnvironmentImpl envImpl) : execution(EnvironmentImpl.new(File, EnvironmentConfig)) && target(envImpl) {
		// before constructor
		envImpl.triggerLatch = LatchSupport.makeSharedLatch("TriggerLatch",
				envImpl);
	}

	/**
	 * Returns the shared trigger latch.
	 */
	public SharedLatch EnvironmentImpl.getTriggerLatch() {
		return triggerLatch;
	}

	pointcut latchedWriteMethods(Database db):(
			execution(* Database.addTrigger(..)) ||
			execution(* Database.removeTrigger(..)) ||
			execution(* Database.removeAllTriggers(..)) 
		) && target(db);

	pointcut latchedReadMethods(Database db):(
			execution(* Database.getSecondaryDatabases(..)) ||
			execution(* Database.notifyTriggers(..))
		) && target(db);

	/**
	 * Log the map tree root and save the LSN.
	 */
	Object around(Database db) throws DatabaseException : 
		latchedWriteMethods(db){
		SharedLatch triggerLatch = db.envHandle.getEnvironmentImpl()
				.getTriggerLatch();
		triggerLatch.acquireExclusive();
		try {
			return proceed(db);
		} finally {
			if (db.triggerList.size() == 0) {
				db.triggerList = null;
			}
			triggerLatch.release();
		}
	}

	Object around(Database db) throws DatabaseException : 
		latchedReadMethods(db) {
		if (db.hasTriggers()) {

			SharedLatch triggerLatch = db.envHandle.getEnvironmentImpl()
					.getTriggerLatch();
			triggerLatch.acquireShared();
			if (db.triggerList == null) {
				db.triggerList = new ArrayList();
			}

			try {
				return proceed(db);
			} finally {
				triggerLatch.release();
			}
		} else
			return proceed(db);
	}

}
