package com.sleepycat.je.log;

import java.util.LinkedList;

import latch.Latch;
import latch.LatchSupport;

import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.dbi.EnvironmentImpl;

public privileged aspect LogBufferPoolExtension {

	/*
	 * bufferPoolLatch is synchronizes access and changes to the buffer pool.
	 * Related latches are the log write latch in LogManager and the read
	 * latches in each log buffer. The log write latch is always taken before
	 * the bufferPoolLatch. The bufferPoolLatch is always taken before any
	 * logBuffer read latch. When faulting in an object from the log, the order
	 * of latching is: bufferPoolLatch.acquire() LogBuffer read latch acquire();
	 * bufferPoolLatch.release(); LogBuffer read latch release() bufferPoolLatch
	 * is also used to protect assignment to the currentWriteBuffer field.
	 */
	// protected not allowed -> needed to make public
	public Latch LogBufferPool.bufferPoolLatch;

	Object around(LogBufferPool lbp) throws DatabaseException:
		(
			call(* hook_resetPool(LinkedList, int)) || 
			call(* hook_allocateBuffer(int))
		) && this(lbp) {
		/*
		 * The following applies when this method is called to reset the pool
		 * when an existing pool is in use: - The old pool will no longer be
		 * referenced. - Buffers being read in the old pool will be no longer
		 * referenced after the read operation is complete. - The
		 * currentWriteBuffer field is not changed here; it will be no longer
		 * referenced after it is written to the file and a new
		 * currentWriteBuffer is assigned. - The logBufferSize can be changed
		 * now because it is only used for allocating new buffers; it is not
		 * used as the size of the currentWriteBuffer.
		 */
		lbp.bufferPoolLatch.acquire();
		Object r = proceed(lbp);
		lbp.bufferPoolLatch.release();
		return r;
	}

	before(LogBufferPool lbp, EnvironmentImpl env) : 
		execution(LogBufferPool.new(FileManager, EnvironmentImpl)) && this(lbp) && args(FileManager, env) {
		lbp.bufferPoolLatch = LatchSupport.makeLatch(LogBufferPool.DEBUG_NAME
				+ "_FullLatch", env);
	}

	Object around(LogBufferPool lb) throws DatabaseException :
		execution(LogBuffer LogBufferPool.hook_getNextToUse())
		&& target(lb) {
		try {
			lb.bufferPoolLatch.acquire();
			return proceed(lb);
		} finally {
			lb.bufferPoolLatch.releaseIfOwner();
		}
	}

	Object around(LogBufferPool lb) throws DatabaseException :
		execution(LogBuffer LogBufferPool.getReadBuffer(long))
		&& target(lb) {
		lb.bufferPoolLatch.acquire();
		try {
			return proceed(lb);
		} finally {
			lb.bufferPoolLatch.releaseIfOwner();
		}
	}

	before(LogBufferPool logBufferPool) throws DatabaseException:
		 execution(void LogBufferPool.writeBufferToFile(int)) && this(logBufferPool) {
		logBufferPool.currentWriteBuffer.latchForWrite();
	}
}
