package cursor;

import latch.LatchNotHeldException;
import latch.LatchSupport;

import com.sleepycat.je.Cursor;
import com.sleepycat.je.DatabaseEntry;
import com.sleepycat.je.DatabaseException;
import com.sleepycat.je.OperationStatus;
import com.sleepycat.je.cleaner.UtilizationProfile;
import com.sleepycat.je.dbi.CursorImpl;
import com.sleepycat.je.dbi.DbTree;
import com.sleepycat.je.dbi.CursorImpl.SearchMode;
import com.sleepycat.je.tree.BIN;
import com.sleepycat.je.tree.DBIN;
import com.sleepycat.je.tree.DIN;
import com.sleepycat.je.tree.DupCountLN;
import com.sleepycat.je.tree.IN;
import com.sleepycat.je.tree.Tree;
import com.sleepycat.je.txn.LockType;
import com.sleepycat.je.txn.Locker;

public privileged aspect CursorImplExtension {

	public BIN CursorImpl.latchBIN() throws DatabaseException {

		while (bin != null) {
			BIN waitingOn = bin;
			waitingOn.latch();
			if (bin == waitingOn) {
				return bin;
			}
			waitingOn.releaseLatch();
		}

		return null;
	}

	public void CursorImpl.releaseBIN() throws LatchNotHeldException {

		if (bin != null) {
			bin.releaseLatchIfOwner();
		}
	}

	public void CursorImpl.latchBINs() throws DatabaseException {

		latchBIN();
		latchDBIN();
	}

	public void CursorImpl.releaseBINs() throws LatchNotHeldException {

		releaseBIN();
		releaseDBIN();
	}

	public DBIN CursorImpl.latchDBIN() throws DatabaseException {

		while (dupBin != null) {
			BIN waitingOn = dupBin;
			waitingOn.latch();
			if (dupBin == waitingOn) {
				return dupBin;
			}
			waitingOn.releaseLatch();
		}
		return null;
	}

	public void CursorImpl.releaseDBIN() throws LatchNotHeldException {

		if (dupBin != null) {
			dupBin.releaseLatchIfOwner();
		}
	}

	/* Check that the target bin is latched. For use in assertions. */
	private boolean CursorImpl.checkAlreadyLatched(boolean alreadyLatched) {
		if (alreadyLatched) {
			if (dupBin != null) {
				return dupBin.isLatchOwner();
			} else if (bin != null) {
				return bin.isLatchOwner();
			}
		}
		return true;
	}

	Object around(CursorImpl ci) throws DatabaseException :
		execution(CursorImpl CursorImpl.hook_cloneCursor(boolean, CursorImpl)) && this(ci) {

		if (ci.nonCloning)
			return proceed(ci);

		try {
			ci.latchBINs();
			return proceed(ci);
		} finally {
			ci.releaseBINs();
		}
	}

	pointcut latchBIN(CursorImpl cursorImpl): (
			execution(int CursorImpl.hook_count(LockType)) 
		) && this(cursorImpl);

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		latchBIN(cursorImpl){
		try {
			cursorImpl.latchBIN();
			return proceed(cursorImpl);
		} finally {
			cursorImpl.releaseBIN();
		}
	}

	pointcut latchBINs(CursorImpl cursorImpl): (
			execution(* CursorImpl.evict(..)) ||
			execution(* CursorImpl.hook_putCurrentInternal(..))
		) && this(cursorImpl);

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		latchBINs(cursorImpl){
		try {
			cursorImpl.latchBINs();
			return proceed(cursorImpl);
		} finally {
			cursorImpl.releaseBINs();
		}
	}

	Object around(DIN dupRoot, CursorImpl cursorImpl) throws DatabaseException :
		execution(DupCountLN CursorImpl.hook_fetchDuplicate(DIN)) &&
		args(dupRoot) && this(cursorImpl) 
	{
		/* Latch couple down the tree. */
		dupRoot.latch();
		cursorImpl.releaseBIN();

		Object r = proceed(dupRoot, cursorImpl);

		/* We can't hold latches when we acquire locks. */
		dupRoot.releaseLatch();

		return r;
	}

	void around(CursorImpl ci) throws DatabaseException : 
		execution(void CursorImpl.removeCursorBIN()) && this(ci) {
		// replace because cannot inject
		BIN abin = ci.latchBIN();
		if (abin != null) {
			abin.removeCursor(ci);
			abin.releaseLatch();
		}
	}

	void around(CursorImpl ci) throws DatabaseException : 
			execution(void CursorImpl.removeCursorDBIN()) && this(ci) {
		// replace because cannot inject
		DBIN abin = ci.latchDBIN();
		if (abin != null) {
			abin.removeCursor(ci);
			abin.releaseLatch();
		}
	}

	OperationStatus around(CursorImpl cursorImpl) throws DatabaseException :
		execution(OperationStatus CursorImpl.delete()) && this(cursorImpl) {
		OperationStatus status = proceed(cursorImpl);
		if (status == OperationStatus.KEYEMPTY)
			cursorImpl.releaseBINs();
		return status;
	}

	after(CursorImpl cursorImpl) throws DatabaseException:
		execution(void CursorImpl.hook_releaseBIN()) && this(cursorImpl){
		cursorImpl.releaseBIN();
	}

	after(CursorImpl cursorImpl) throws DatabaseException:
		execution(void CursorImpl.hook_releaseBINs()) && this(cursorImpl){
		cursorImpl.releaseBINs();
	}

	after(CursorImpl cursorImpl) throws DatabaseException:
		execution(void CursorImpl.hook_latchBINs()) && this(cursorImpl){
		cursorImpl.latchBINs();
	}

	after(IN targetIN) throws DatabaseException:
		execution(void CursorImpl.hook_releaseLatchIfOwner(IN)) && args(targetIN){
		targetIN.releaseLatchIfOwner();
	}

	after(IN targetIN) throws DatabaseException:
		execution(void CursorImpl.hook_releaseLatch(IN)) && args(targetIN){
		targetIN.releaseLatch();
	}

	after(IN targetIN) throws DatabaseException:
		execution(void CursorImpl.hook_latch(IN)) && args(targetIN){
		targetIN.latch();
	}

	after(CursorImpl cursorImpl) throws DatabaseException:
		execution(void CursorImpl.hook_releaseDBIN()) && this(cursorImpl){
		cursorImpl.releaseDBIN();
	}

	boolean around(CursorImpl cursorImpl) throws DatabaseException :
		execution(boolean CursorImpl.hook_lookNextKeyForInsertInner(DatabaseEntry, DatabaseEntry, SearchMode, boolean)) 
		&& this(cursorImpl) 
	{
		boolean latched = true;
		try {
			latched = proceed(cursorImpl);
		} finally {
			if (latched) {
				cursorImpl.releaseBINs();
			}
		}
		return latched;

	}

	pointcut assertNoLatches() : 
		execution(* CursorImpl.putLN(..)) ||
		execution(* CursorImpl.hook_assertNoLatches());

	before():assertNoLatches(){
		assert LatchSupport.countLatchesHeld() == 0 : LatchSupport.latchesHeldToString();
	}

	pointcut assertIsLatchOwner(CursorImpl cursorImpl) : 
		execution(* CursorImpl.getLatchedDupRoot(..)) &&
		this(cursorImpl);

	before(CursorImpl cursorImpl) : assertIsLatchOwner(cursorImpl){
		assert cursorImpl.bin.isLatchOwner();
	}

	Object around(CursorImpl cursorImpl, boolean wait) throws DatabaseException : 
		call(* Locker.lock(..)) && withincode(* CursorImpl.lockLNDeletedAllowed(..)) && this(cursorImpl) && if (wait==false)&& args(*,*,wait,*) {
		cursorImpl.releaseBINs();
		Object r = proceed(cursorImpl, wait);
		cursorImpl.latchBINs();
		return r;
	}

	Object around(CursorImpl cursorImpl, boolean wait, DIN dupRoot)
			throws DatabaseException : 
		call(* Locker.lock(..)) && this(cursorImpl) &&  args(*,*,wait,*) && 
			cflowbelow(execution(* CursorImpl.lockDupCountLN(..)) && args(dupRoot,*))  {
		/* Release all latches. */
		if (wait==false) {
			dupRoot.releaseLatch();
			cursorImpl.releaseBINs();
		}
		return proceed(cursorImpl, wait, dupRoot);
	}

	Object around(CursorImpl cursorImpl, DIN dupRoot) throws DatabaseException : 
		call(* IN.fetchTarget(..)) && this(cursorImpl) &&
			cflowbelow(execution(* CursorImpl.lockDupCountLN(..)) && args(dupRoot,*))  {
		/* Reacquire all latches. */
		cursorImpl.latchBIN();
		Object r = proceed(cursorImpl, dupRoot);
		dupRoot.latch();
		cursorImpl.latchDBIN();
		return r;
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		call(* CursorImpl.getDupRoot(..)) && target(cursorImpl) &&
			cflowbelow(execution(* CursorImpl.delete(..)))  {

		return cursorImpl.getLatchedDupRoot(true);
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		call(* CursorImpl.getDupRoot(..)) && target(cursorImpl) &&
			!cflowbelow(execution(* CursorImpl.delete(..)))  {

		return cursorImpl.getLatchedDupRoot(true);
	}

	public DIN CursorImpl.getLatchedDupRoot(boolean isDBINLatched)
			throws DatabaseException {

		assert bin != null;
		assert index >= 0;

		DIN dupRoot = (DIN) bin.fetchTarget(index);

		if (isDBINLatched) {

			/*
			 * The BIN and DBIN are currently latched and we need to latch the
			 * dupRoot, which is between the BIN and DBIN in the tree. First
			 * trying latching the dupRoot no-wait; if this works, we have
			 * latched out of order, but in a way that does not cause deadlocks.
			 * If we don't get the no-wait latch, then release the DBIN latch
			 * and latch in the proper order down the tree.
			 */
			if (!dupRoot.latchNoWait()) {
				releaseDBIN();
				dupRoot.latch();
				latchDBIN();
			}
		} else {
			dupRoot.latch();
		}

		return dupRoot;
	}

	pointcut getCurrentAlreadyLatchedSources() :
			withincode(* Cursor.positionAllowPhantoms(..)) ||
			withincode(* Cursor.searchInternal(..)) ||
			within(UtilizationProfile) ||
			withincode(* CursorImpl.getNextDuplicate(..)) ||
			withincode(* CursorImpl.getNextWithKeyChangeStatus(..)) ||
			within(DbTree);

//	OperationStatus around(CursorImpl cursorImpl, DatabaseEntry foundKey,
//			DatabaseEntry foundData, LockType lockType)
//			throws DatabaseException :
//		call(* CursorImpl.getCurrent(DatabaseEntry,DatabaseEntry,LockType)) && 
//		args(foundKey,foundData,lockType) && getCurrentAlreadyLatchedSources() &&
//		this(cursorImpl) 
//	{
//		assert cursorImpl.checkAlreadyLatched(true) : cursorImpl.dumpToString(true);
//		return cursorImpl.getCurrentAlreadyLatched(foundKey, foundData,
//				lockType, true);
//	}

	OperationStatus around(CursorImpl cursorImpl, DatabaseEntry foundKey,
			DatabaseEntry foundData, LockType lockType, boolean first)
			throws DatabaseException :
		call(* CursorImpl.getCurrent(DatabaseEntry,DatabaseEntry,LockType, boolean)) && 
		args(foundKey,foundData,lockType, first) && getCurrentAlreadyLatchedSources() &&
		this(cursorImpl) {
		assert cursorImpl.checkAlreadyLatched(true) : cursorImpl.dumpToString(true);
		return cursorImpl.getCurrentAlreadyLatched(foundKey, foundData,
				lockType, first);
	}

	/**
	 * Retrieve the current record. Assume the bin is already latched. Return
	 * with the target bin unlatched.
	 */
	public OperationStatus CursorImpl.getCurrentAlreadyLatched(
			DatabaseEntry foundKey, DatabaseEntry foundData, LockType lockType,
			boolean first) throws DatabaseException {

		assert assertCursorState(true) : dumpToString(true);
		assert checkAlreadyLatched(true) : dumpToString(true);

		if (bin == null) {
			return OperationStatus.KEYEMPTY;
		}

		try {
			return fetchCurrent(foundKey, foundData, lockType, first);
		} finally {
			releaseBINs();
		}
	}

	before(CursorImpl cursorImpl) throws DatabaseException:
		 call(* CursorImpl.fetchCurrent(..)) && this(cursorImpl)&&
		 withincode(* CursorImpl.getCurrent(..)) {
		if (cursorImpl.dupBin == null) {
			cursorImpl.latchBIN();
		} else {
			cursorImpl.latchDBIN();
		}
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException:
		 call(* CursorImpl.fetchCurrent(..)) && target(cursorImpl) &&
		 withincode(* CursorImpl.getCurrent(..)) {

		assert cursorImpl.checkAlreadyLatched(true) : cursorImpl.dumpToString(true);
		try {
			return proceed(cursorImpl);
		} finally {
			cursorImpl.releaseBINs();
		}
	}

	pointcut getCurrentLNAlreadyLatchedSources() :
		within(DbTree) && !withincode(* DbTree.getDbName(..));

	before(CursorImpl cursorImpl) throws DatabaseException:
		 call(* CursorImpl.getCurrentLN(..)) && target(cursorImpl) &&
		 !getCurrentLNAlreadyLatchedSources() {
		if (cursorImpl.bin != null) {
			cursorImpl.latchBIN();
		}
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException:
			 call(* CursorImpl.hook_getCurrentLNInternal(..)) && target(cursorImpl) {

		try {
			assert cursorImpl.checkAlreadyLatched(true) : cursorImpl.dumpToString(true);
			return proceed(cursorImpl);
		} finally {
			cursorImpl.releaseBINs();
		}
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		execution(* CursorImpl.flushBINToBeRemoved()) && this(cursorImpl) {
		cursorImpl.binToBeRemoved.latch();
		Object r = proceed(cursorImpl);
		cursorImpl.binToBeRemoved.releaseLatch();
		return r;
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		execution(* CursorImpl.flushDBINToBeRemoved()) && this(cursorImpl) {
		cursorImpl.dupBinToBeRemoved.latch();
		Object r = proceed(cursorImpl);
		cursorImpl.dupBinToBeRemoved.releaseLatch();
		return r;
	}

	pointcut releaseBINsAfterwards(CursorImpl cursorImpl) :
		target(cursorImpl) && (
			call(* CursorImpl.hook_fetchCurrentInternal(..))				
		);

	Object around(CursorImpl cursorImpl) throws DatabaseException:
		releaseBINsAfterwards(cursorImpl) {
		try {
			return proceed(cursorImpl);
		} finally {
			cursorImpl.releaseBINs();
		}
	}

	before(BIN bin):execution(* CursorImpl.addCursor(..)) && args(bin) {
		if (bin != null) {
			assert bin.isLatchOwner();
		}
	}

	void around(CursorImpl cursorImpl) throws DatabaseException :
		call(* CursorImpl.removeCursorDBIN()) && this(cursorImpl) && 
		withincode(* CursorImpl.clearDupBIN(..)) && 
		cflowbelow(execution(* Tree.insert(..))) {

		cursorImpl.dupBin.removeCursor(cursorImpl);
		cursorImpl.dupBin.releaseLatch();
	}

	Object around(CursorImpl cursorImpl) throws DatabaseException : 
		execution(* CursorImpl.searchAndPosition(..)) && target(cursorImpl) {
		try {
			return proceed(cursorImpl);
		} catch (DatabaseException e) {
			/* Release latch on error. */
			cursorImpl.releaseBIN();
			throw e;
		}

	}

	before(CursorImpl cursorImpl, DIN duplicateRoot) throws DatabaseException :
		call(* Tree.searchSubTree(..)) && withincode(* CursorImpl.searchAndPositionBoth(..))&& 
		args(duplicateRoot,..) && 
		this(cursorImpl) {
		duplicateRoot.latch();
		cursorImpl.releaseBIN();
	}

	before(CursorImpl cursorImpl) throws DatabaseException :
		call(* CursorImpl.hook_fetchCurrentStart()) && this(cursorImpl){
		assert cursorImpl.targetBin.isLatchOwner();
	}
}
